// Game.cpp: implementation of the cGame class and children.
//
//////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "gamedambuilder.h"
#include "critterarmed.h"
#include "critterwall.h"
#include "force.h"
#include "listener.h"
#include "pop.h" //for playSound
#include "popview.h"
#include "spritebubble.h"
#include "critterviewer.h"
#include "graphicsopengl.h"
#include "graphicsmfc.h"
#include "spritequake.h"

IMPLEMENT_SERIAL( cGameDambuilder, cGame, 0);
IMPLEMENT_SERIAL( cCritterDamFloat, cCritter, 0);
IMPLEMENT_SERIAL( cCritterDamPlayer, cCritterArmedPlayer, 0);
IMPLEMENT_SERIAL( cCritterDamWall, cCritterWall, 0);
IMPLEMENT_SERIAL( cCritterDamPlayerBullet, cCritterBullet, 0);

//#define DAMPLAYERGRAVITY
	/* If we define DAMPLAYERGRAVITYin, then the player will be subject to gravity.
	 You will only notice this with the ListenerCar or ListenerSpaceship control. 
	 Car won't look good, as it will be confusing when the car is falling backwards and
	it thinks it's in reverse. */

//#define PROPGRAVITY
	/* If you want to turn off gravity for the props & bu, turn off PROPGRAVITY. */

//#define DEBUGDAM
	/* Turn on DEBUGDAM to turn off all but one of the dam walls and all the floats. */

//#define STARTIN2D
	/* Turn on STARTIN2D to start in 2D, otherwise start in 3D. */
//#define SPRITEQUAKE
	/* Use Quake sprites for the critters. Only do this if you plan to start in 3D. */
#define WEAKBOUNCE
	/* By default I set the player's absorberflag to TRUE so he doesn't bounce
	at all.  But I can bounce just a little by setting absorberflag to FALSE
	and _bounciness to, say, 0.2 */

Real cGameDambuilder::WALLTHICKNESS = 0.4;
Real cGameDambuilder::PLAYERRADIUS = 0.5;
Real cGameDambuilder::MAXPLAYERSPEED = 30.0;
Real cGameDambuilder::WALLPRISMDZ = 0.3; //Use this as the average height of walls.
Real cGameDambuilder::DAMBACKGROUNDOFFSET = 0.01; //Amount to lower the floor.
Real cGameDambuilder::DAMPLAYERBOUNCINESS = 0.05; /* Let's have the player
	bounce only a small amount off the walls and other players. */
//==============================cGameDambuilder=======
cGameDambuilder::cGameDambuilder()
{
	_menuflags &= ~cGame::MENU_SHIELD;
		/* Default _menuflags from cGame has hopper off, autoplay off.
			We don't use the shield control in the standard game as nobody is
			shooting at us. */
#ifdef DAMPLAYERGRAVITY
	_menuflags |= cGame::MENU_HOPPER; 
		// The Hopper may be interesting if player has gravity.
#endif //DAMPLAYERGRAVITY
//Pick the cursor tools.
	//_arrayHCURSOR has only the _hCursorArrow inherited from cGame.
	_arrayHCURSOR.Add(((CPopApp*)::AfxGetApp())->_hCursorPin);
	_arrayHCURSOR.Add(((CPopApp*)::AfxGetApp())->_hCursorDragger);
	_arrayHCURSOR.Add(((CPopApp*)::AfxGetApp())->_hCursorReplicate);
//Initialize other fields
	_border.set( 25, 35) ; 
	cGame::BACKGROUNDOFFSET = cGameDambuilder::DAMBACKGROUNDOFFSET;
//Set the backgroudn bitmap
	setBackgroundBitmap(IDB_BACKGROUND); /*  Sets the cSpriteBackground for the background bitmap.
		Note that you must call this after the _border has its size fully initialized. So each
		cGame child with a different border should call this again in its constructor. */
	setWrapflag(cCritter::WRAP);
	_spritetype = cGame::ST_MESHSKIN;  //cGame::ST_SIMPLEPOLYGONS;
	_seedcount = cGame::COUNTLARGE;
#ifdef SPRITEQUAKE
	_spritetype = cGame::ST_MESHSKIN;
	_seedcount = cGame::COUNTSMALL; //Will run slow on some cards with many cSpriteQuake
#endif //SPRITEQUAKE
//Set the player 
	setPlayer(new cCritterDamPlayer());
	pplayer()->moveTo(cVector(_border.locorner() + cVector(1,1,0)));
	pplayer()->lookAt(cVector::ZEROVECTOR);
//Put in some walls in the constructor.
#ifndef DEBUGDAM
	new cCritterDamWall( cVector(-4, 2),	cVector(3, -2), this);
	new cCritterDamWall( cVector(1.5, -2),	cVector(6, 0.3),this);
	new cCritterDamWall( cVector(-5, 0),	cVector(-5, 9), this);
	new cCritterDamWall( cVector(8, -2),	cVector(12, -3), this);
	new cCritterDamWall( cVector(-8, -4),	cVector(-6, -3), this);
	new cCritterDamWall( cVector(-9, -8),	cVector(-3, -3), this);
#else //DEBUGDAM is true  Just have one wall.
	new cCritterDamWall( cVector(-4, 2),	cVector(4, 2), this);
	_seedcount = 0;
#endif //DEBUGDAM
}

void cGameDambuilder::seedCritters()
{
 	_pbiota->purgeNonPlayerNonWallCritters(); /* Clean out all critters but leave 
		player and	walls. */
	for (int i=0; i < _seedcount; i++)
		new cCritterDamFloat(this);
}

void cGameDambuilder::initializeView(CPopView *pview)
{
	cGame::initializeView(pview); 
		/* Good idea to call the baseclass method or you
		might leave out something important that it does,
		possibly resetting something that was changed
		by a different games. */
	pview->setCursor(((CPopApp*)::AfxGetApp())->_hCursorDragger);
	pview->setUseBackgroundBitmap(TRUE);
	//pview->setUseSolidBackground(FALSE);
	CPoint panepos = pview->paneRowCol();
	if (panepos == CPoint(0,0))
	{
		pview->setGraphicsClass(RUNTIME_CLASS(cGraphicsOpenGL));
		pview->pviewpointcritter()->setListener(new cListenerViewerRide());
	}
	if (panepos == CPoint(0,1)) //Use flat map for second view.
	{
		pview->setGraphicsClass(RUNTIME_CLASS(cGraphicsMFC)); 
		pview->setUseBackgroundBitmap(FALSE);
	}
}

void cGameDambuilder::initializeViewpoint(cCritterViewer *pviewer)
{
	cGame::initializeViewpoint(pviewer); /* No overload.  Note that you don't actually
		care about this call if you plan to ride the player, as then the viewpoint is
		set by the player. */
	pviewer->zoom(0.5);
}

void cGameDambuilder::drawBackground(cGraphics *pgraphics, int drawflags)
{
	cGame::drawBackground(pgraphics, drawflags);
}

//=====================cCritterDamFloat===========

cCritterDamFloat::cCritterDamFloat(cGame *pownergame):
cCritter(pownergame)
{
	if (pgame())
		setSprite(pgame()->randomSprite(pgame()->spritetype()));
	randomize(cCritter::MF_POSITION | cCritter::MF_VELOCITY); 
	addForce(new cForceEvadeBullet());
#ifdef PROPGRAVITY
	addForce(new cForceGravity()); //Uses default gravity strength 25.
	addForce(new cForceDrag()); //Uses default friction strength 0.5.
#endif //PROPGRAVITY
	setRadius(0.4);
	setDensity(2.0); 
	setMaxspeed(30.0);
}

//===================cCritterDamPlayer================
cCritterDamPlayer::cCritterDamPlayer(cGame *pownergame):
cCritterArmedPlayer(pownergame)
{
	setListener(new cListenerCar());
 	setBulletClass(RUNTIME_CLASS(cCritterDamPlayerBullet));
	setWaitShoot(2.0*cCritterArmed::WAITSHOOT); //Don't shoot as fast as usual.
	setMaxspeed(cGameDambuilder::MAXPLAYERSPEED);
	setListenerAcceleration(160.0); //So Hopper can overcome gravity
	setRadius(cGameDambuilder::PLAYERRADIUS);
	setPrismDz(0.4);
	setDensity(2.0); 
	addForce(new cForceDrag(1.5)); /* Point of using friction is so he'll
		slow down.  If you make him denser to knock other critters around
		more you may need to turn up friction, it can be higher than 1.0,
		though if you make it way high critter acts weird, with
		"friction" actually reversing its direction. */
#ifndef WEAKBOUNCE //No bounce at all off walls or other critters
	setAbsorberflag(TRUE); //Keeps player from being buffeted about.
#else //WEAKBOUNCE
	setAbsorberflag(FALSE); 
	setBounciness(cGameDambuilder::DAMPLAYERBOUNCINESS); //Try like 0.2
#endif //WEAKBOUNCE
#ifdef DAMPLAYERGRAVITY
	setListenerAcceleration(80.0); //So Car or Spaceship can overcome gravity
	addForce(new cForceGravity()); //Uses default gravity strength 25.
	addForce(new cForceDrag()); //Uses default friction strength 0.5.
#endif //DAMPLAYERGRAVITY
}

//===================cCritterDamWall
cCritterDamWall::cCritterDamWall(const cVector &enda, const cVector &endb,
	 Real nthickness, Real nheight, cGame *pownergame):
cCritterWall(enda, endb, nthickness, nheight, pownergame)
{
	psprite()->setFillColor(cRandomizer::pinstance()->randomColor());
	_movebox.setZRange(-100.0, 100.0); /* So you can comfortably position the 
		3D box so its sprite appears in a nice place, even though your game is
		really using a 2D world. Might be better just to use a 3D _border
		in the first place. */
	moveTo(cVector(_position.x(), _position.y(), height()/2.0));
		/*Let's try and align all the walls with base on the xy plane.*/
}

cCritterDamWall::cCritterDamWall(const cVector &enda, const cVector &endb,
	cGame *pownergame):
cCritterWall(enda, endb, cGameDambuilder::WALLTHICKNESS, cGameDambuilder::WALLPRISMDZ, pownergame)
{
	setHeight(cRandomizer::pinstance()->randomReal(0.2, 3.0) *
		cGameDambuilder::WALLPRISMDZ);
	psprite()->setFillColor(cRandomizer::pinstance()->randomColor());
	_movebox.setZRange(-100.0, 100.0); /* So you can comfortably position the 
		3D box so its sprite appears in a nice place, even though your game is
		really using a 2D world. Might be better just to use a 3D _border
		in the first place. */
	_dragbox.setZRange(-100.0, 100.0); /* Also need to fix this so that when
		we drag, it doesn't get clamped down. */
	moveTo(cVector(_position.x(), _position.y(), height()/2.0));
		/*Let's try and align all the walls with base on the xy plane.*/
}

//====================cCritterDamPlayerBullet==================
cCritterDamPlayerBullet::cCritterDamPlayerBullet()
{
	setRadius(2.0*cCritter::BULLETRADIUS); //Make bigger than usual.
}

void cCritterDamPlayerBullet::initialize(cCritterArmed *pshooter)
{
	cCritterBullet::initialize(pshooter);
	clearForcearray();/* The cCritterBullet::initialize copies the physics 
		forces from pshooter to the bullet, and if I don't want the car's
		friction, I have to turn it off here by clearing the force array. */
}

cCritterDamPlayerBullet::isTarget(cCritter *pcritter)
{
	return cCritterBullet::isTarget(pcritter) && 
		pcritter->GetRuntimeClass() != RUNTIME_CLASS(cCritterDamWall);
}
