#ifndef GAMESTUB
#define GAMESTUB

/*
	The gamestub.* files are meant to be a possible starting point for your games.  Alter the
	gamestub.* files freely.  
	In the code below, we put the word virtual in front of all the methods that we have
	inherited from the base classes and which we plan to overload.  (These methods were
	already declared as virtual in the base classes, so that our overloads can take effect.
	Declaring them virtual in this child class only means that they could be further overloaded 
	in the children of the child.)
*/

#include "game.h"
#include "critterarmed.h"

//=== cGameStub Critters ====================================

class cCritterStubPlayer : public cCritterArmedPlayer //Our player.
{
DECLARE_SERIAL(cCritterStubPlayer);
public:
	cCritterStubPlayer(cGame *pownergame = NULL);
//overloads
	virtual void reset();
	virtual int damage(int hitstrength);
	virtual BOOL collide(cCritter *pcritter); 
	virtual void update(CPopView *pactiveview, Real dt);
	virtual cCritterBullet* shoot();
//	virtual void setWrapflag(int wrapflag){}//Don't allow _wrapflag to change from cCritter::BOUNCE.
};

class cCritterStubPlayerBullet : public cCritterBullet
{
DECLARE_SERIAL(cCritterStubPlayerBullet);
public:
	cCritterStubPlayerBullet();
	virtual void update(CPopView *pactiveview, Real dt);
};

class cCritterStubRival : public cCritterArmedRobot //Or just a cCritter if not shooting
{
DECLARE_SERIAL(cCritterStubRival);
public:
	cCritterStubRival(cGame *pownergame = NULL); /* The defualt arg
		means this defines a no-arg constructor as well.*/
//overloads
	virtual int damage(int hitstrength);
	virtual BOOL collide(cCritter *pcritter); 
	virtual void update(CPopView *pactiveview, Real dt);
};

	/* The cCritterBulletSilver only can damage the ptarget of the
	shooter.   */
class cCritterStubRivalBullet : public cCritterBulletSilver
{
DECLARE_SERIAL(cCritterStubRivalBullet);
public:
	cCritterStubRivalBullet();
	virtual void initialize(cCritterArmed *pshooter);
};

class cCritterStubProp : public cCritter
{
DECLARE_SERIAL(cCritterStubProp);
public:
	cCritterStubProp(cGame *pownergame = NULL); 
//overloads
	virtual int damage(int hitstrength);
	virtual BOOL collide(cCritter *pcritter); 
	virtual void update(CPopView *pactiveview, Real dt);
};

//======================= cGameStub ================


class cGameStub : public cGame
{
DECLARE_SERIAL(cGameStub);
//Name your statics here
static int PLAYERHEALTH;
static int DEFAULTRIVALCOUNT;
static int DEFAULTSEEDCOUNT;
private:
	int _rivalcount; //possible param, don't have to use it.
public:
	cGameStub();
 //overloads
	virtual void Serialize(CArchive& ar); //Overload for _rivalcount
	virtual void reset();
	virtual void adjustGameParameters();
	virtual CString statusMessage();
	virtual void initializeView(CPopView *pview);
	virtual void initializeViewpoint(cCritterViewer *pviewer); //To change the default view direction.
	virtual void seedCritters();
};

#endif //GAMESTUB